from itertools import chain
from typing import Optional, Tuple

from ._base import prepare
from .background import Background
from .location import LocationMixin
from .scenario import Scenario
from .tag import Tag


@prepare
class RuleChildren:
    background: Optional[Background] = None
    scenario: Optional[Scenario] = None

    def __iter__(self):
        if self.background is not None:
            yield from self.background

        if self.scenario is not None:
            yield from self.scenario


@prepare
class Rule(LocationMixin):
    keyword: str
    name: str
    tags: Tuple[Tag, ...]
    children: Tuple[RuleChildren, ...]
    description: str

    def __iter__(self):
        yield from self.tags

        yield self

        yield from chain.from_iterable(self.children)



from pathlib import Path

import click
from attr import dataclass

from .utils import err, out


@dataclass
class Report:
    """Provides a reformatting counter. Can be rendered with `str(report)`."""

    check: bool
    change_count: int = 0
    same_count: int = 0
    failure_count: int = 0

    def done(self, path: str, changed: bool) -> None:
        """Increment the counter for successful reformatting. Write out a message."""
        if changed:
            reformatted = "Would reformat" if self.check else "Reformatted"
            out(f"{reformatted} {path}")

            self.change_count += 1
        else:
            self.same_count += 1

    def failed(self, path: Path, message: str) -> None:
        """Increment the counter for failed reformatting. Write out a message."""
        err(f"Error: cannot format {path}: {message}")
        self.failure_count += 1

    @property
    def return_code(self) -> int:
        """Return the exit code that the app should use.

        This considers the current state of changed files and failures:
        - if there were any failures, return 123;
        - if any files were changed and --check is being used, return 1;
        - otherwise return 0.
        """
        # According to http://tldp.org/LDP/abs/html/exitcodes.html starting with
        # 126 we have special return codes reserved by the shell.
        if self.failure_count:
            return 123

        elif self.change_count and self.check:
            return 1

        return 0

    def __str__(self) -> str:
        """Render a color report of the current state.

        Use `click.unstyle` to remove colors.
        """
        if self.check:
            reformatted = "would be reformatted"
            unchanged = "would be left unchanged"
            failed = "would fail to reformat"
        else:
            reformatted = "reformatted"
            unchanged = "left unchanged"
            failed = "failed to reformat"

        report_lines = []
        if self.change_count:
            s = "s" if self.change_count > 1 else ""
            report_lines.append(
                click.style(f"{self.change_count} file{s} {reformatted}", bold=True)
            )
        if self.same_count:
            s = "s" if self.same_count > 1 else ""
            report_lines.append(f"{self.same_count} file{s} {unchanged}")
        if self.failure_count:
            s = "s" if self.failure_count > 1 else ""
            report_lines.append(
                click.style(f"{self.failure_count} file{s} {failed}", fg="red")
            )
        return ", ".join(report_lines) + "."



from itertools import chain
from typing import Tuple

from ._base import prepare
from .examples import Examples
from .location import LocationMixin
from .step import Step
from .tag import Tag


@prepare
class Scenario(LocationMixin):
    keyword: str
    name: str
    steps: Tuple[Step, ...]
    tags: Tuple[Tag, ...]
    description: str
    examples: Tuple[Examples, ...]

    def __iter__(self):
        yield from self.tags

        yield self

        yield from chain.from_iterable(self.steps)

        yield from chain.from_iterable(self.examples)



from attr import attrib

from ._base import prepare
from .location import LocationMixin


def escape_table_cell_value(text: str) -> str:
    r"""
    Escape special characters in a table cell's value. There are three of them:
        - \\
        - \|
        - \n
    (Source: https://github.com/cucumber/common/blob/7cdd5259c90410971877dbe480733ba1b44e9a62/gherkin/testdata/good/escaped_pipes.feature)
    """

    return text.replace("\\", "\\\\").replace("|", "\\|").replace("\n", "\\n")


@prepare
class TableCell(LocationMixin):
    value: str = attrib(converter=escape_table_cell_value)



from itertools import chain
from typing import Optional, Tuple

from ._base import prepare
from .location import LocationMixin
from .rule import Rule, RuleChildren
from .tag import Tag


@prepare
class FeatureChildren(RuleChildren):
    rule: Optional[Rule] = None

    def __iter__(self):
        yield from super().__iter__()

        if self.rule is not None:
            yield from self.rule


@prepare
class Feature(LocationMixin):
    language: str
    keyword: str
    name: str
    children: Tuple[FeatureChildren, ...]
    tags: Tuple[Tag, ...]
    description: str

    def __iter__(self):
        yield from self.tags

        yield self

        yield from chain.from_iterable(self.children)



from typing import Optional

from ._base import prepare
from .data_table import DataTable
from .doc_string import DocString
from .location import LocationMixin


@prepare
class Step(LocationMixin):
    keyword: str
    text: str
    data_table: Optional[DataTable] = None
    doc_string: Optional[DocString] = None

    def __iter__(self):
        yield self

        if self.doc_string is not None:
            yield from self.doc_string

        if self.data_table is not None:
            yield from self.data_table
